/*
 * Copyright (c) 2010-2013 Michael Kuhn
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define _POSIX_C_SOURCE 200809L

#include <glib.h>
#include <glib-unix.h>

#include <mpi.h>

#include <math.h>
#include <pthread.h>
#include <string.h>

struct BenchmarkResult
{
	gdouble read;
	gdouble write;
};

typedef struct BenchmarkResult BenchmarkResult;

struct BenchmarkStatistics
{
	gdouble read_sum;
	gdouble read_sum2;
	gdouble write_sum;
	gdouble write_sum2;
};

typedef struct BenchmarkStatistics BenchmarkStatistics;

struct ThreadData
{
	guint id;
	BenchmarkResult result;
};

typedef struct ThreadData ThreadData;

static gboolean opt_mpi = FALSE;
static gboolean opt_posix = FALSE;

static gint opt_block_count = 1000;
static gint opt_block_size = 4096;
static gboolean opt_sync = FALSE;
static gboolean opt_sync_on_close = FALSE;
static gint opt_repetitions = 1;
static gint opt_threads = 1;
static gchar* opt_command = NULL;
static gchar* opt_path = NULL;

static gboolean opt_mpi_atomic = FALSE;

static gint process_id = 0;
static gint process_num = 1;

static pthread_barrier_t thread_barrier;

static
guint64
get_num_nodes (void)
{
	static guint64 num_nodes = 0;

	gchar const* env_vars[] = {
		"OMPI_MCA_orte_num_nodes",
		"SLURM_JOB_NUM_NODES"
	};

	gchar const* env_num_nodes;

	if (num_nodes != 0)
	{
		return num_nodes;
	}

	for (guint i = 0; i < G_N_ELEMENTS(env_vars); i++)
	{
		env_num_nodes = g_getenv(env_vars[i]);

		if (env_num_nodes != NULL)
		{
			num_nodes = g_ascii_strtoull(env_num_nodes, NULL, 10);
		}

		if (num_nodes != 0)
		{
			break;
		}
	}

	return num_nodes;
}

static
void
print_statistics (BenchmarkResult* result, BenchmarkStatistics* statistics)
{
	static gboolean printed_header = FALSE;

	guint64 bytes;

	gdouble write_thp;
	gdouble read_thp;
	guint64 num_nodes;

	if (result != NULL)
	{
		gdouble read;
		gdouble write;

		MPI_Reduce(&(result->read), &read, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
		MPI_Reduce(&(result->write), &write, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

		result->read = read / process_num;
		result->write = write / process_num;
	}

	if (process_id != 0)
	{
		return;
	}

	if (result == NULL)
	{
		if (opt_repetitions == 1)
		{
			return;
		}

		g_print("# configuration[n/p] write_mean[MB/s] write_stddev[MB/s] read_mean[MB/s] read_stddev[MB/s]\n");

		if ((num_nodes = get_num_nodes()) != 0)
		{
			g_print("%" G_GUINT64_FORMAT "/", num_nodes);
		}

		g_print("%d ", process_num * opt_threads);
		g_print("%.3f ", statistics->write_sum / opt_repetitions);
		g_print("%.3f ", sqrt(((opt_repetitions * statistics->write_sum2) - (statistics->write_sum * statistics->write_sum)) / (opt_repetitions * (opt_repetitions - 1))));
		g_print("%.3f ", statistics->read_sum / opt_repetitions);
		g_print("%.3f\n", sqrt(((opt_repetitions * statistics->read_sum2) - (statistics->read_sum * statistics->read_sum)) / (opt_repetitions * (opt_repetitions - 1))));

		return;
	}

	result->read /= opt_threads;
	result->write /= opt_threads;

	bytes = (guint64)opt_block_size * opt_block_count * process_num * opt_threads;

	//g_print("Size:  %s * %d * %d * %d = %s\n", size_str, opt_block_count, process_num, opt_threads, total_size_str);
	if (!printed_header)
	{
		g_print("# configuration[n/p] write_time[s] write_throughput[MB/s] read_time[s] read_throughput[MB/s]\n");
		printed_header = TRUE;
	}

	if ((num_nodes = get_num_nodes()) != 0)
	{
		g_print("%" G_GUINT64_FORMAT "/", num_nodes);
	}

	write_thp = bytes / result->write / (1024 * 1024);
	read_thp = bytes / result->read / (1024 * 1024);

	g_print("%d ", process_num * opt_threads);
	g_print("%.3f ", result->write);
	g_print("%.3f ", write_thp);
	g_print("%.3f ", result->read);
	g_print("%.3f\n", read_thp);

	statistics->read_sum += read_thp;
	statistics->read_sum2 += read_thp * read_thp;
	statistics->write_sum += write_thp;
	statistics->write_sum2 += write_thp * write_thp;
}

static
guint64
get_offset (ThreadData* thread_data, guint64 iteration)
{
	return opt_block_size * ((iteration * process_num * opt_threads) + (process_id * opt_threads) + thread_data->id);
}

static
gboolean
manage_file (ThreadData* thread_data)
{
	return (process_id == 0 && thread_data->id == 0);
}

static
void
barrier (ThreadData* thread_data)
{
	if (thread_data->id == 0)
	{
		MPI_Barrier(MPI_COMM_WORLD);
	}

	pthread_barrier_wait(&thread_barrier);
}

static
gchar*
get_name (ThreadData* thread_data)
{
	return g_strdup("small-access");
}

static
gpointer
small_access_mpi (gpointer data)
{
	ThreadData* thread_data = data;

	gchar buf[opt_block_size * 10];

	MPI_Comm comm;
	MPI_File file[1];
	MPI_Status status[1];
	GTimer* timer;
	gchar* name;
	gchar* path;
	guint64 iteration;
	gint ret;

	comm = MPI_COMM_WORLD;
	name = get_name(thread_data);
	path = g_build_filename(opt_path, name, NULL);
	g_free(name);

	MPI_File_open(comm, path, MPI_MODE_RDWR | MPI_MODE_CREATE, MPI_INFO_NULL, file);
	MPI_File_close(file);

	barrier(thread_data);

	timer = g_timer_new();
	ret = MPI_File_open(comm, path, MPI_MODE_RDWR, MPI_INFO_NULL, file);

	if (ret != MPI_SUCCESS)
	{
		g_error("Process %d can not write.\n", process_id);
	}

	if (opt_mpi_atomic)
	{
		MPI_File_set_atomicity(file[0], 1);
	}

	barrier(thread_data);

	for (iteration = 0; iteration < (guint)opt_block_count; iteration += 10)
	{
		guint remaining;

		remaining = MIN(10, opt_block_count - iteration);

		for (guint i = 0; i < remaining; i++)
		{
			gint bytes;

			MPI_File_write_at(file[0], get_offset(thread_data, iteration + i), buf + opt_block_size * i, opt_block_size, MPI_CHAR, status);
			MPI_Get_count(status, MPI_CHAR, &bytes);
			g_assert(bytes == opt_block_size);

			if (opt_sync)
			{
				MPI_File_sync(file[0]);
			}
		}

		barrier(thread_data);
	}

	if (opt_sync_on_close)
	{
		MPI_File_sync(file[0]);
	}

	MPI_File_close(file);
	thread_data->result.write = g_timer_elapsed(timer, NULL);

	barrier(thread_data);

	if (opt_command != NULL && thread_data->id == 0)
	{
		system(opt_command);
	}

	barrier(thread_data);

	g_timer_start(timer);
	ret = MPI_File_open(comm, path, MPI_MODE_RDWR, MPI_INFO_NULL, file);

	if (ret != MPI_SUCCESS)
	{
		g_error("Process %d can not read.\n", process_id);
	}

	if (opt_mpi_atomic)
	{
		MPI_File_set_atomicity(file[0], 1);
	}

	barrier(thread_data);

	for (iteration = 0; iteration < (guint)opt_block_count; iteration += 10)
	{
		guint remaining;

		remaining = MIN(10, opt_block_count - iteration);

		for (guint i = 0; i < remaining; i++)
		{
			gint bytes;

			MPI_File_read_at(file[0], get_offset(thread_data, iteration + i), buf + opt_block_size * i, opt_block_size, MPI_CHAR, status);
			MPI_Get_count(status, MPI_CHAR, &bytes);
			g_assert(bytes == opt_block_size);
		}

		barrier(thread_data);
	}

	MPI_File_close(file);

	thread_data->result.read = g_timer_elapsed(timer, NULL);
	g_timer_destroy(timer);

	MPI_File_delete(path, MPI_INFO_NULL);

	g_free(path);

	return thread_data;
}

static
gpointer
small_access_posix (gpointer data)
{
	ThreadData* thread_data = data;

	gchar buf[opt_block_size * 10];

	GTimer* timer;
	gchar* name;
	gchar* path;
	gint fd;
	guint64 iteration = 0;

	name = get_name(thread_data);
	path = g_build_filename(opt_path, name, NULL);
	timer = g_timer_new();
	g_free(name);

	barrier(thread_data);

	if (manage_file(thread_data))
	{
		fd = open(path, O_RDWR | O_CREAT, 0600);

		if (fd == -1)
		{
			g_printerr("Process %d cannot create file on path %s\n", process_id, path);
		}

		close(fd);
	}

	barrier(thread_data);

	g_timer_start(timer);
	fd = open(path, O_RDWR);

	if (fd == -1)
	{
		//g_error("%s\n",strerror(errno));
		g_error("Process %d can not write on %s.\n", process_id, path);
	}

	for (iteration = 0; iteration < (guint)opt_block_count; iteration += 10)
	{
		guint remaining;

		remaining = MIN(10, opt_block_count - iteration);

		for (guint i = 0; i < remaining; i++)
		{
			guint64 bytes;

			bytes = pwrite(fd, buf + opt_block_size * i, opt_block_size, get_offset(thread_data, iteration + i));
			g_assert(bytes == (guint)opt_block_size);

			if (opt_sync)
			{
				fsync(fd);
			}
		}

		barrier(thread_data);
	}

	if (opt_sync_on_close)
	{
		fsync(fd);
	}

	close(fd);
	thread_data->result.write = g_timer_elapsed(timer, NULL);

	barrier(thread_data);

	if (opt_command != NULL && thread_data->id == 0)
	{
		system(opt_command);
	}

	barrier(thread_data);

	g_timer_start(timer);
	fd = open(path, O_RDWR);

	if (fd == -1)
	{
		g_error("Process %d can not read.\n", process_id);
	}

	//barrier(thread_data);

	for (iteration = 0; iteration < (guint)opt_block_count; iteration += 10)
	{
		guint remaining;

		remaining = MIN(10, opt_block_count - iteration);

		for (guint i = 0; i < remaining; i++)
		{
			guint64 bytes;

			bytes = pread(fd, buf + opt_block_size * i, opt_block_size, get_offset(thread_data, iteration + i));
			g_assert(bytes == (guint)opt_block_size);
		}

		barrier(thread_data);
	}

	close(fd);
	thread_data->result.read = g_timer_elapsed(timer, NULL);

	g_timer_destroy(timer);

	if (manage_file(thread_data))
	{
		unlink(path);
	}

	g_free(path);

	return thread_data;
}

int
main (int argc, char** argv)
{
	GError* error = NULL;
	GOptionContext* context;
	GOptionGroup* group;

	GOptionEntry entries[] = {
		{ "posix", 0, 0, G_OPTION_ARG_NONE, &opt_posix, "Use POSIX I/O", NULL },
		{ "mpi", 0, 0, G_OPTION_ARG_NONE, &opt_mpi, "Use MPI-I/O", NULL },
		{ "block-count", 0, 0, G_OPTION_ARG_INT, &opt_block_count, "Block count", "1000" },
		{ "block-size", 0, 0, G_OPTION_ARG_INT, &opt_block_size, "Block size", "4096" },
		{ "sync",  0, 0, G_OPTION_ARG_NONE, &opt_sync, "Use synchronisation", NULL },
		{ "sync-on-close",  0, 0, G_OPTION_ARG_NONE, &opt_sync_on_close, "Use synchronisation on close", NULL },
		{ "repetitions", 0, 0, G_OPTION_ARG_INT, &opt_repetitions, "Number of repetitions to perform", "1" },
		{ "threads", 0, 0, G_OPTION_ARG_INT, &opt_threads, "Number of threads to use", "1" },
		{ "command", 0, 0, G_OPTION_ARG_STRING, &opt_command, "Command to execute between write and read phases", NULL },
		{ "path", 0, 0, G_OPTION_ARG_STRING, &opt_path, "Path to use", NULL },
		{ NULL, 0, 0, 0, NULL, NULL, NULL }
	};

	GOptionEntry entries_mpi[] = {
		{ "mpi-atomic", 0, 0, G_OPTION_ARG_NONE, &opt_mpi_atomic, "Use atomic mode", NULL },
		{ NULL, 0, 0, 0, NULL, NULL, NULL }
	};

	BenchmarkStatistics statistics;
	GThread** threads;
	GThreadFunc thread_func = NULL;

	{
		gint thread_provided;

		MPI_Init_thread(&argc, &argv, MPI_THREAD_SERIALIZED, &thread_provided);

		g_assert(thread_provided == MPI_THREAD_SERIALIZED);

		MPI_Comm_rank(MPI_COMM_WORLD, &process_id);
		MPI_Comm_size(MPI_COMM_WORLD, &process_num);
	}

	context = g_option_context_new(NULL);
	g_option_context_add_main_entries(context, entries, NULL);

	group = g_option_group_new("mpi", "MPI-I/O Options", "Show MPI-I/O help options", NULL, NULL);
	g_option_group_add_entries(group, entries_mpi);
	g_option_context_add_group(context, group);

	if (!g_option_context_parse(context, &argc, &argv, &error))
	{
		g_option_context_free(context);

		if (error)
		{
			g_printerr("%s\n", error->message);
			g_error_free(error);
		}

		return 1;
	}

	if ((!opt_mpi && !opt_posix) || opt_path == NULL)
	{
		gchar* help;

		help = g_option_context_get_help(context, TRUE, NULL);
		g_option_context_free(context);

		g_print("%s", help);
		g_free(help);

		return 1;
	}

	g_option_context_free(context);

	pthread_barrier_init(&thread_barrier, NULL, opt_threads);
	threads = g_new(GThread*, opt_threads);

	if (opt_posix)
	{
		thread_func = small_access_posix;
	}
	else if (opt_mpi)
	{
		g_assert(opt_threads == 1);
		thread_func = small_access_mpi;
	}

	g_assert(thread_func != NULL);

	statistics.read_sum = 0.0;
	statistics.read_sum2 = 0.0;
	statistics.write_sum = 0.0;
	statistics.write_sum2 = 0.0;

	for (gint i = 0; i < opt_repetitions; i++)
	{
		BenchmarkResult global_result;

		global_result.read = 0.0;
		global_result.write = 0.0;

		for (gint j = 0; j < opt_threads; j++)
		{
			ThreadData* thread_data;

			thread_data = g_slice_new(ThreadData);
			thread_data->id = j;

			threads[j] = g_thread_new("small-access", thread_func, thread_data);
		}

		for (gint j = 0; j < opt_threads; j++)
		{
			ThreadData* thread_data;

			thread_data = g_thread_join(threads[j]);

			global_result.read += thread_data->result.read;
			global_result.write += thread_data->result.write;

			g_slice_free(ThreadData, thread_data);
		}

		print_statistics(&global_result, &statistics);
		// FIXME sleep?
	}

	print_statistics(NULL, &statistics);

	pthread_barrier_destroy(&thread_barrier);

	g_free(opt_command);

	MPI_Finalize();

	return 0;
}
